program GeneratorsTest;

{$APPTYPE CONSOLE}

uses
  Types,
  SysUtils,
  Generics.Collections,
  Generators in 'Generators.pas';

// Simple predicates

function IsOdd(I: Integer): Boolean;
begin
  Result := I mod 2 = 1;
end;

function IsPrime(Number: Integer): Boolean;
var
  i: Integer;
begin
  if Number = 1 then
    Exit(False);
  for i := 2 to Number div 2 do
    if Number mod i = 0 then
      Exit(False);
  Exit(True);
end;

// Parameterized predicates

function LessThan(Max: Integer): TPredicate<Integer>;
begin
  Result :=
    function(Number: Integer): Boolean
    begin
      Result := Number < Max;
    end;
end;

// Mapping functions

function Square(Number: Integer): Integer;
begin
  Result := Number * Number;
end;

// Folding functions

function Add(Number1, Number2: Integer): Integer;
begin
  Result := Number1 + Number2;
end;

function Mult(Number1, Number2: Integer): Integer;
begin
  Result := Number1 * Number2;
end;

begin
  try
    { "Nothing" happens - a TIntegersGenerator object is created,
      and destroyed again when it goes out of scope }
    Ints(5, 10, 1);

    WriteLn('----- Simple list generators -----');
    WriteLn('');

    WriteLn('All integers between 1 and 10');
    WriteLn(Ints(1, 1, 10).ToStr(IntToStr));
    WriteLn('');

    WriteLn('All odd integers between 1 and 10');
    WriteLn(Ints(1, 2, 10).ToStr(IntToStr));
    WriteLn('');

    WriteLn('All odd integers between 1 and 10, using a filter');
    WriteLn(Ints(1, 1, 10).Filter(IsOdd).ToStr(IntToStr));
    WriteLn('');

    WriteLn('The first 100 prime numbers, using take. Note that');
    WriteLn('we use an infinite list by omitting MaxValue');
    WriteLn(Ints(1, 1).Filter(IsPrime).Take(100).ToStr(IntToStr));
    WriteLn('');

    WriteLn('All primes less than 100, using TakeWhile and a');
    WriteLn('parameterized predicate. Note that we have omitted the');
    WriteLn('StartValue and Step parameters - they both default to 1.');
    WriteLn(Ints.Filter(IsPrime).TakeWhile(LessThan(100)).ToStr(IntToStr));
    WriteLn('');

    WriteLn('All primes between 100 and 200, using DropWhile and TakeWhile.');
    WriteLn(Ints.Filter(IsPrime).DropWhile(LessThan(100)).TakeWhile(LessThan(200)).ToStr(IntToStr));
    WriteLn('');

    WriteLn('The squares of the first 10 prime numbers, using map.');
    WriteLn(Ints.Filter(IsPrime).Map(Square).Take(10).ToStr(IntToStr));
    WriteLn('');

    WriteLn('----- Simple list generators - aggregation -----');
    WriteLn('');

    WriteLn('The sum of (1..10), introducing Fold.');
    WriteLn(Ints(1, 1, 10).Fold(Add, 0));
    WriteLn('');

    WriteLn('The factorial of 10');
    WriteLn(Ints(1, 1, 10).Fold(Mult, 1));
    WriteLn('');

    WriteLn('The sum of all primes less than 1000.');
    WriteLn(Ints.Filter(IsPrime).TakeWhile(LessThan(1000)).Fold(Add, 0));
    WriteLn('');

    ReadLn;
  except
    on E: Exception do
      Writeln(E.ClassName, ': ', E.Message);
  end;

end.
